/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifndef __GNOME_SEARCH_PROVIDER2_H__
#define __GNOME_SEARCH_PROVIDER2_H__

#include <gio/gio.h>

G_BEGIN_DECLS


/* ------------------------------------------------------------------------ */
/* Declarations for org.gnome.Shell.SearchProvider2 */

#define SHELL_TYPE_SEARCH_PROVIDER2 (shell_search_provider2_get_type ())
#define SHELL_SEARCH_PROVIDER2(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), SHELL_TYPE_SEARCH_PROVIDER2, ShellSearchProvider2))
#define SHELL_IS_SEARCH_PROVIDER2(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), SHELL_TYPE_SEARCH_PROVIDER2))
#define SHELL_SEARCH_PROVIDER2_GET_IFACE(o) (G_TYPE_INSTANCE_GET_INTERFACE ((o), SHELL_TYPE_SEARCH_PROVIDER2, ShellSearchProvider2Iface))

struct _ShellSearchProvider2;
typedef struct _ShellSearchProvider2 ShellSearchProvider2;
typedef struct _ShellSearchProvider2Iface ShellSearchProvider2Iface;

struct _ShellSearchProvider2Iface
{
  GTypeInterface parent_iface;

  gboolean (*handle_activate_result) (
    ShellSearchProvider2 *object,
    GDBusMethodInvocation *invocation,
    const gchar *arg_identifier,
    const gchar *const *arg_terms,
    guint arg_timestamp);

  gboolean (*handle_get_initial_result_set) (
    ShellSearchProvider2 *object,
    GDBusMethodInvocation *invocation,
    const gchar *const *arg_terms);

  gboolean (*handle_get_result_metas) (
    ShellSearchProvider2 *object,
    GDBusMethodInvocation *invocation,
    const gchar *const *arg_identifiers);

  gboolean (*handle_get_subsearch_result_set) (
    ShellSearchProvider2 *object,
    GDBusMethodInvocation *invocation,
    const gchar *const *arg_previous_results,
    const gchar *const *arg_terms);

  gboolean (*handle_launch_search) (
    ShellSearchProvider2 *object,
    GDBusMethodInvocation *invocation,
    const gchar *const *arg_terms,
    guint arg_timestamp);

};

GType shell_search_provider2_get_type (void) G_GNUC_CONST;

GDBusInterfaceInfo *shell_search_provider2_interface_info (void);
guint shell_search_provider2_override_properties (GObjectClass *klass, guint property_id_begin);


/* D-Bus method call completion functions: */
void shell_search_provider2_complete_get_initial_result_set (
    ShellSearchProvider2 *object,
    GDBusMethodInvocation *invocation,
    const gchar *const *results);

void shell_search_provider2_complete_get_subsearch_result_set (
    ShellSearchProvider2 *object,
    GDBusMethodInvocation *invocation,
    const gchar *const *results);

void shell_search_provider2_complete_get_result_metas (
    ShellSearchProvider2 *object,
    GDBusMethodInvocation *invocation,
    GVariant *metas);

void shell_search_provider2_complete_activate_result (
    ShellSearchProvider2 *object,
    GDBusMethodInvocation *invocation);

void shell_search_provider2_complete_launch_search (
    ShellSearchProvider2 *object,
    GDBusMethodInvocation *invocation);



/* D-Bus method calls: */
void shell_search_provider2_call_get_initial_result_set (
    ShellSearchProvider2 *proxy,
    const gchar *const *arg_terms,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean shell_search_provider2_call_get_initial_result_set_finish (
    ShellSearchProvider2 *proxy,
    gchar *** out_results,
    GAsyncResult *res,
    GError **error);

gboolean shell_search_provider2_call_get_initial_result_set_sync (
    ShellSearchProvider2 *proxy,
    const gchar *const *arg_terms,
    gchar *** out_results,
    GCancellable *cancellable,
    GError **error);

void shell_search_provider2_call_get_subsearch_result_set (
    ShellSearchProvider2 *proxy,
    const gchar *const *arg_previous_results,
    const gchar *const *arg_terms,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean shell_search_provider2_call_get_subsearch_result_set_finish (
    ShellSearchProvider2 *proxy,
    gchar *** out_results,
    GAsyncResult *res,
    GError **error);

gboolean shell_search_provider2_call_get_subsearch_result_set_sync (
    ShellSearchProvider2 *proxy,
    const gchar *const *arg_previous_results,
    const gchar *const *arg_terms,
    gchar *** out_results,
    GCancellable *cancellable,
    GError **error);

void shell_search_provider2_call_get_result_metas (
    ShellSearchProvider2 *proxy,
    const gchar *const *arg_identifiers,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean shell_search_provider2_call_get_result_metas_finish (
    ShellSearchProvider2 *proxy,
    GVariant ** out_metas,
    GAsyncResult *res,
    GError **error);

gboolean shell_search_provider2_call_get_result_metas_sync (
    ShellSearchProvider2 *proxy,
    const gchar *const *arg_identifiers,
    GVariant ** out_metas,
    GCancellable *cancellable,
    GError **error);

void shell_search_provider2_call_activate_result (
    ShellSearchProvider2 *proxy,
    const gchar *arg_identifier,
    const gchar *const *arg_terms,
    guint arg_timestamp,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean shell_search_provider2_call_activate_result_finish (
    ShellSearchProvider2 *proxy,
    GAsyncResult *res,
    GError **error);

gboolean shell_search_provider2_call_activate_result_sync (
    ShellSearchProvider2 *proxy,
    const gchar *arg_identifier,
    const gchar *const *arg_terms,
    guint arg_timestamp,
    GCancellable *cancellable,
    GError **error);

void shell_search_provider2_call_launch_search (
    ShellSearchProvider2 *proxy,
    const gchar *const *arg_terms,
    guint arg_timestamp,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean shell_search_provider2_call_launch_search_finish (
    ShellSearchProvider2 *proxy,
    GAsyncResult *res,
    GError **error);

gboolean shell_search_provider2_call_launch_search_sync (
    ShellSearchProvider2 *proxy,
    const gchar *const *arg_terms,
    guint arg_timestamp,
    GCancellable *cancellable,
    GError **error);



/* ---- */

#define SHELL_TYPE_SEARCH_PROVIDER2_PROXY (shell_search_provider2_proxy_get_type ())
#define SHELL_SEARCH_PROVIDER2_PROXY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), SHELL_TYPE_SEARCH_PROVIDER2_PROXY, ShellSearchProvider2Proxy))
#define SHELL_SEARCH_PROVIDER2_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), SHELL_TYPE_SEARCH_PROVIDER2_PROXY, ShellSearchProvider2ProxyClass))
#define SHELL_SEARCH_PROVIDER2_PROXY_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), SHELL_TYPE_SEARCH_PROVIDER2_PROXY, ShellSearchProvider2ProxyClass))
#define SHELL_IS_SEARCH_PROVIDER2_PROXY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), SHELL_TYPE_SEARCH_PROVIDER2_PROXY))
#define SHELL_IS_SEARCH_PROVIDER2_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), SHELL_TYPE_SEARCH_PROVIDER2_PROXY))

typedef struct _ShellSearchProvider2Proxy ShellSearchProvider2Proxy;
typedef struct _ShellSearchProvider2ProxyClass ShellSearchProvider2ProxyClass;
typedef struct _ShellSearchProvider2ProxyPrivate ShellSearchProvider2ProxyPrivate;

struct _ShellSearchProvider2Proxy
{
  /*< private >*/
  GDBusProxy parent_instance;
  ShellSearchProvider2ProxyPrivate *priv;
};

struct _ShellSearchProvider2ProxyClass
{
  GDBusProxyClass parent_class;
};

GType shell_search_provider2_proxy_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (ShellSearchProvider2Proxy, g_object_unref)
#endif

void shell_search_provider2_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
ShellSearchProvider2 *shell_search_provider2_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error);
ShellSearchProvider2 *shell_search_provider2_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);

void shell_search_provider2_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
ShellSearchProvider2 *shell_search_provider2_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error);
ShellSearchProvider2 *shell_search_provider2_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);


/* ---- */

#define SHELL_TYPE_SEARCH_PROVIDER2_SKELETON (shell_search_provider2_skeleton_get_type ())
#define SHELL_SEARCH_PROVIDER2_SKELETON(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), SHELL_TYPE_SEARCH_PROVIDER2_SKELETON, ShellSearchProvider2Skeleton))
#define SHELL_SEARCH_PROVIDER2_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), SHELL_TYPE_SEARCH_PROVIDER2_SKELETON, ShellSearchProvider2SkeletonClass))
#define SHELL_SEARCH_PROVIDER2_SKELETON_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), SHELL_TYPE_SEARCH_PROVIDER2_SKELETON, ShellSearchProvider2SkeletonClass))
#define SHELL_IS_SEARCH_PROVIDER2_SKELETON(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), SHELL_TYPE_SEARCH_PROVIDER2_SKELETON))
#define SHELL_IS_SEARCH_PROVIDER2_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), SHELL_TYPE_SEARCH_PROVIDER2_SKELETON))

typedef struct _ShellSearchProvider2Skeleton ShellSearchProvider2Skeleton;
typedef struct _ShellSearchProvider2SkeletonClass ShellSearchProvider2SkeletonClass;
typedef struct _ShellSearchProvider2SkeletonPrivate ShellSearchProvider2SkeletonPrivate;

struct _ShellSearchProvider2Skeleton
{
  /*< private >*/
  GDBusInterfaceSkeleton parent_instance;
  ShellSearchProvider2SkeletonPrivate *priv;
};

struct _ShellSearchProvider2SkeletonClass
{
  GDBusInterfaceSkeletonClass parent_class;
};

GType shell_search_provider2_skeleton_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (ShellSearchProvider2Skeleton, g_object_unref)
#endif

ShellSearchProvider2 *shell_search_provider2_skeleton_new (void);


G_END_DECLS

#endif /* __GNOME_SEARCH_PROVIDER2_H__ */
